#ifndef DEF_BEHAVIOURPROCESSOR_H
#define DEF_BEHAVIOURPROCESSOR_H

class BehaviourProcessor;

class IBehaviour {
public:
	virtual bool Update() = 0; //returns true if behaviour is finished
	virtual void SetBehaviourProcessor(BehaviourProcessor* b) = 0;
};

class BehaviourProcessor {
public:
	float* dampingForceFactor;
	float* vibrationAmplitude;
	Vec3f* currentPos;
	vector<IBehaviour*> behaviours;

	void AddBehaviour(IBehaviour* b) {
		b->SetBehaviourProcessor(this);
		behaviours.push_back(b);
	}

	bool UpdateBehaviours() {

		if (behaviours.size()>0 && behaviours[0]->Update()) {
			behaviours.erase(behaviours.begin());
		}
		return behaviours.size()!=0;
	}


	bool MoveTo(Vec3f& targetPos, float dampingForceFactor = 10000.0f, float vibrationAmplitude = 1.0f, float stepSize = 0.02f) {
		*(this->dampingForceFactor) = dampingForceFactor;
		*(this->vibrationAmplitude) = vibrationAmplitude;

		Vec3f d = targetPos - *(this->currentPos);
		float magnitude = d.length();
		if (stepSize < magnitude) {
			d.normalize();
			*(this->currentPos) = *(this->currentPos) + d * stepSize;
		}
		else {
			*(this->currentPos) = *(this->currentPos) + d;
		}

		if ((*(this->currentPos) - targetPos).sqrLength() < 0.001f) {
			return true;
		}
		return false;
	}
};

#endif
