#ifndef DEF_MODEL_H
#define DEF_MODEL_H

#include "Vec3.h"
#include "SceneObjects.h"
#include "BVHierarchy.h"
#include <map>
#include <vector>

using namespace std;

class Model {

private:
	// this is a map, where the keys are ids and the values are triangles
	map<int, Triangle*> triangleMap;
	vector<SceneObject*> objects;
	BVHierarchyNode* BVHroot;
	bool isBuild = false;

public:
	vector<Triangle*> triangles;

	//! constructor
	Model() 
	{

	}

	//! destructor
	~Model() {
	
	}

	void setTriangle(int id, float ax, float ay, float az, float bx, float by, float bz, float cx, float cy, float cz) {
		Triangle* t;
		if (triangleMap.count(id) == 0) {
			triangleMap[id] = new Triangle();
		}
		else {
			t = triangleMap[id];
		}

		t = triangleMap[id];

		t->a.x = ax;
		t->a.y = ay;
		t->a.z = az;

		t->b.x = bx;
		t->b.y = by;
		t->b.z = bz;

		t->c.x = cx;
		t->c.y = cy;
		t->c.z = cz;
	}

	void removeTriangle(int id) {
		/*Triangle* t = triangleMap[id]; //Todo
		if(t)
			triangles.erase((SceneObject*)t);*/
		triangleMap.erase(id);
	}

	void clear() {
		triangles.clear();
		triangleMap.clear();
		objects.clear();
		isBuild = false;
	}

	void build() {
		triangles.clear();
		objects.clear();
		for (pair<int, Triangle*> element : triangleMap) {
			triangles.push_back(element.second);
			objects.push_back(element.second);
			
		}
		BVHroot = new BVHierarchyNode(objects);
		isBuild = true;
	}

	bool getBuildStatus() {
		return isBuild;
	}

	BoundingBox getBoundingBox() {
		if (!isBuild)
			return BoundingBox();
		return BVHroot->getBoundingBox();
	}

	bool getClosestPointInBox(Vec3f& p, BoundingBox& box, float& closestDistance, Vec3f& closestPoint) const
	{
		if (!isBuild)
			return false;
		closestDistance = FLT_MAX;
		return BVHroot->getClosestPointInBoxDepthSearch(p, box, closestDistance, closestPoint);
	}

	bool getAdjacentEdge(Vec3f& a, Vec3f& b, Triangle* parentTriangle, Triangle*& resultParentTriangle) const{
		if (!isBuild)
			return false;
		return BVHroot->getAdjacentEdge(a,b,parentTriangle,resultParentTriangle);
	}

	//returns true if intersection was found, false otherwise
	bool getIntersection(const Ray& ray,RayHit& hit) const
	{
		if (!isBuild)
			return false;
		return BVHroot->intersect(ray,hit);
	}

	Vec3f getClosestPoint(Vec3f p, float& distance) {
		float minDistance = FLT_MAX, currentDistance;
		Vec3f minP, currentP;
		
		for (pair<int, Triangle*> element : triangleMap) {
			currentP = element.second->getClosestPoint(p);
			currentDistance = (p - currentP).length();
			if (currentDistance < minDistance) {
				minDistance = currentDistance;
				minP = currentP;
			}
		}

		distance = minDistance;
		return minP;
	}
};

#endif