#ifndef DEF_SPATIALGRID_H
#define DEF_SPATIALGRID_H

#include "Vec3.h"
#include <algorithm>
#include <set>
#include <math.h>
#include <algorithm>

using namespace std;

template<typename T> class SpatialGrid {

private:
	template<typename T> class GridCell {
	private:
		int arraySize = 16;
	
	public:
		int objectCount;
		T* objects;

		GridCell<T>() {
			init();
		}

		~GridCell<T>() {
			delete[] objects;
		}

		void init() {
			objectCount = 0;
			objects = new T[arraySize];
		}

		void doubleArraySize() {
			T* newArray = new T[arraySize * 2];
			for (int i = 0; i < arraySize; ++i) {
				newArray[i] = objects[i];
			}
			arraySize *= 2;
			delete[] objects;
			objects = newArray;
		}

		void insert(T object) {
			if (objectCount >= arraySize) {
				doubleArraySize();
			}
			objects[objectCount] = object;
			++objectCount;
		}

		void remove(T object) {
			int index;
			for (int i = 0; i < objectCount; ++i) {
				if (objects[i] == object) {
					index = i;
					break;
				}
			}
			for (int i = index; i < objectCount - 1; ++i) {
				objects[i] = objects[i + 1];
			}
		}

		void clear() {
			objectCount = 0;
		}
	};

	unsigned int width,height,depth;
	float cellSize;
	unsigned long int size;
	GridCell<T>* cells;

	unsigned long int getIndex(Vec3f p) {
		unsigned int x = int(floor(p.x / cellSize))%width;
		unsigned int y = int(floor(p.y/ cellSize))%height;
		unsigned int z = int(floor(p.z / cellSize))%depth;
		return x + y * width + z * width*height;
	}

public:
	SpatialGrid(unsigned int _width, unsigned int _height, unsigned int _depth, float _cellSize) {

		size = _width * _height*_depth;
		cells = new GridCell<T>[size];
		cellSize = _cellSize;
		width = _width;
		height = _height;
		depth = _depth;
	}

	~SpatialGrid() {
		delete[] cells;
	}

	void get(Vec3f p, set<T>& result) {
		unsigned long int index = getIndex(p);
		GridCell<T>* c = &cells[index];

		for (int i = 0; i < c->objectCount; ++i) {
			result.insert(c->objects[i]);
		}
	}

	void get(Vec3f min, Vec3f max, set<T>& result) {
		Vec3f i;
		for (i.z = min.z; i.z <= max.z + cellSize; i.z += cellSize) {
			for (i.y = min.y; i.y <= max.y + cellSize; i.y += cellSize) {
				for (i.x = min.x; i.x <= max.x + cellSize; i.x += cellSize) {
					get(i, result);
				}
			}
		}
	}

	void insert(Vec3f min, Vec3f max, T obj) {
		Vec3f i;
		GridCell<T>* c;
		for (i.z = min.z; i.z <= max.z+ cellSize; i.z += cellSize) {
			for (i.y = min.y; i.y <= max.y + cellSize; i.y += cellSize) {
				for (i.x = min.x; i.x <= max.x + cellSize; i.x += cellSize) {
					c = &cells[getIndex(i)];
					c->insert(obj);
				}
			}
		}
	}

	void remove(Vec3f min, Vec3f max, T obj) {
		Vec3f i;
		GridCell<T>* c;
		for (i.z = min.z; i.z <= max.z + cellSize; i.z += cellSize) {
			for (i.y = min.y; i.y <= max.y + cellSize; i.y += cellSize) {
				for (i.x = min.x; i.x <= max.x + cellSize; i.x += cellSize) {
					c = &cells[getIndex(i)];
					c->remove(obj);
				}
			}
		}
	}

	void clear() {
		for (unsigned long int i = 0; i < size; ++i) {
			cells[i].clear();
		}
	}
};

#endif