#ifndef DEF_TRIANGLE_H
#define DEF_TRIANGLE_H

#include "Vec3.h"
#include <algorithm>

using namespace std;

class Triangle {

private:
	float clamp01(float value) {
		return max(0.0f, min(1.0f, value));
	}

public:
	struct {
		int id;
		Vec3f a;
		Vec3f b;
		Vec3f c;
	};

	//! constructor
	Triangle()
	{
	}

	Triangle(Vec3f va, Vec3f vb, Vec3f vc)
	{
		a = va;
		b = vb;
		c = vc;
	}

	Triangle(float ax,float ay,float az,float bx, float by, float bz, float cx, float cy, float cz)
	{
		a = Vec3f(ax,ay,az);
		b = Vec3f(bx, by, bz);
		c = Vec3f(cx, cy, cz);
	}

	//! destructor
	~Triangle() {}

	Vec3f getClosestPoint(const Vec3f &p)
	{
		Vec3f edge0 = b - a;
		Vec3f edge1 = c - a;
		Vec3f v0 = a - p;

		float af = edge0.dot(edge0);
		float bf = edge0.dot(edge1);
		float cf = edge1.dot(edge1);
		float df = edge0.dot(v0);
		float ef = edge1.dot(v0);

		float det = af * cf - bf * bf;
		float s = bf * ef - cf * df;
		float t = bf * df - af * ef;

		if (s + t < det)
		{
			if (s < 0.f)
			{
				if (t < 0.f)
				{
					if (df < 0.f)
					{
						s = clamp01(-df / af);
						t = 0.f;
					}
					else
					{
						s = 0.f;
						t = clamp01(-ef / cf);
					}
				}
				else
				{
					s = 0.f;
					t = clamp01(-ef / cf);
				}
			}
			else if (t < 0.f)
			{
				s = clamp01(-df / af);
				t = 0.f;
			}
			else
			{
				float invDet = 1.f / det;
				s *= invDet;
				t *= invDet;
			}
		}
		else
		{
			if (s < 0.f)
			{
				float tmp0 = bf + df;
				float tmp1 = cf + ef;
				if (tmp1 > tmp0)
				{
					float numer = tmp1 - tmp0;
					float denom = af - 2 * bf + cf;
					s = clamp01(numer / denom);
					t = 1 - s;
				}
				else
				{
					t = clamp01(-ef / cf);
					s = 0.f;
				}
			}
			else if (t < 0.f)
			{
				if (af + df > bf + ef)
				{
					float numer = cf + ef - bf - df;
					float denom = af - 2 * bf + cf;
					s = clamp01(numer / denom);
					t = 1 - s;
				}
				else
				{
					s = clamp01(-ef / cf);
					t = 0.f;
				}
			}
			else
			{
				float numer = cf + ef - bf - df;
				float denom = af - 2 * bf + cf;
				s = clamp01(numer / denom);
				t = 1.f - s;
			}
		}

		return a + edge0 * s + edge1 * t;
	}

	Vec3f getMin() {
		Vec3f min = a;
		
		if (b.x < min.x)
			min.x = b.x;
		if (b.y < min.y)
			min.y = b.y;
		if (b.z < min.z)
			min.z = b.z;

		if (c.x < min.x)
			min.x = c.x;
		if (c.y < min.y)
			min.y = c.y;
		if (c.z < min.z)
			min.z = c.z;
		
		return min;
	}

	Vec3f getMax() {
		Vec3f max = a;

		if (b.x > max.x)
			max.x = b.x;
		if (b.y > max.y)
			max.y = b.y;
		if (b.z > max.z)
			max.z = b.z;

		if (c.x > max.x)
			max.x = c.x;
		if (c.y > max.y)
			max.y = c.y;
		if (c.z > max.z)
			max.z = c.z;
		return max;
	}
};

#endif