var cmd_getPosition = {cmd : 'getPosition'};
var cmd_getButtons = {cmd : 'getButtons'};
var cmd_ping = {cmd : 'ping'};
var cmd_getConnection = {cmd : 'getConnection'};
var cmd_getCollision = {cmd : 'getCollision'};
var cmd_getClosestPoint = {cmd : 'getClosestPoint'};
var cmd_clearModel = {cmd : 'clearModel'};
var cmd_buildModel = {cmd : 'buildModel'}
var cmd_setTriangle = {cmd : 'setTriangle', id : 0, ax : 0.0, ay : 0.0, az : 0.0, bx : 0.0, by : 0.0, bz : 0.0, cx : 0.0, cy : 0.0, cz : 0.0};
var cmd_setTriangles = {cmd : 'setTriangles', vertexOffset : 0, vertexCount : 0, vertexArray : ''};
var cmd_removeTriangle = {cmd : 'removeTriangle', id : 0};
var cmd_setCoursorSize = {cmd: 'setCoursorSize', size : 0.1};
var cmd_setHapticMode = {cmd: 'setHapticMode', mode : 0};

var cmd_createProjection = {cmd: 'createProjection', px: 0.0, py: 0.0, pz: 1.0};

var position = {cmd : 'getPosition', x : 0.0, y: 0.0, z: 0.0};
var buttons = {cmd : 'getButtons', button0 : 0, button1 : 0, button2 : 0, button3 : 0};
var connection = -1;
var coursorSize = 0.1;
var hapticMode = 0;

var collision = false;
var closestPoint = {cmd : 'getClosestPoint', x : 0.0, y: 0.0, z: 0.0};
var modelSize = 0;
var lastLoad = 0;
var loading  = 0.0;

var cmdArray = [];
  
function pushCommand(cmdObj){
	/*if(connection<0 && cmdObj.cmd!='getConnection'){
		return false;
	}*/
	cmdArray.unshift(cmdObj);
	return true;
}

function processResult(result){
	switch(result.cmd){
		case "debug": //debug and error messages should be handled in the contentScript
			break;
		case "error":
			break;
		case "getPosition":
			position = result;
			break;
		case "getButtons":
			buttons = result;
			break;
		case "getConnection":
			connection = result.connection;
			break;
		case "getCollision":
			collision = result.collision;
			break;
		case "getClosestPoint":
			closestPoint = result;
			break;
		case "setTriangle":
			if(result.id-lastLoad>10000){
				loading = result.id/modelSize;
				lastLoad = result.id;
				percentComplete = loading;
				console.log("loading model: "+Math.floor(100*loading)+"%");
			}
			break;
		case "setTriangles":
			if(result.id-lastLoad>10000){
				loading = result.id/modelSize;
				lastLoad = result.id;
				percentComplete = loading;
				console.log("loading model: "+Math.floor(100*loading)+"%");
			}
			break;
		case "removeTriangle":
			break;
		case "createProjection":
			console.log(result);
			break;
		case "clearModel":
			console.log(result);
			break;
		case "buildModel":
			console.log('model is ready for haptics.');
			break;
	}
}

function processCommandArray(){
	var currentCmd = document.getElementById('fd_command').innerHTML;
	if(currentCmd==""){
		var result = document.getElementById('fd_return').innerHTML;
		if(result != "")
			processResult(JSON.parse(result));
		if(cmdArray.length>0){
			document.getElementById('fd_command').innerHTML = JSON.stringify(cmdArray.pop());
		}
	}
}

function refreshPositionAndButtons(){
	if(connection>=0 && cmdArray.length<10){
		pushCommand(cmd_getPosition);
		pushCommand(cmd_getButtons);
		pushCommand(cmd_getCollision);
		pushCommand(cmd_getClosestPoint);
		cmd_setCoursorSize.size = coursorSize;
		pushCommand(cmd_setCoursorSize);
		cmd_setHapticMode.mode = hapticMode;
		pushCommand(cmd_setHapticMode);
	}
}

setInterval(processCommandArray,0);
setInterval(refreshPositionAndButtons,20);
pushCommand(cmd_getConnection);
pushCommand(cmd_clearModel);




//TIRANGLE AND MODEL SYNCHRONIZATION

function setTriangle(id, ax, ay, az, bx, by, bz, cx, cy, cz){
	cmd_setTriangle.id = id;
	
	cmd_setTriangle.ax = ax;
	cmd_setTriangle.ay = ay;
	cmd_setTriangle.az = az;
	
	cmd_setTriangle.bx = bx;
	cmd_setTriangle.by = by;
	cmd_setTriangle.bz = bz;
	
	cmd_setTriangle.cx = cx;
	cmd_setTriangle.cy = cy;
	cmd_setTriangle.cz = cz;

	pushCommand(JSON.parse(JSON.stringify(cmd_setTriangle))); //deep-copy json parsable object
}


function setTriangles(vertexArray, start, length){
	triangles = new Array(length);
	triangleHexString = "";
	
	for(var i = start; i<start+length;++i)
	{
		triangleHexString+=ToHex(vertexArray[i]);
	}
	
	cmd_setTriangles.vertexArray = triangleHexString;
	cmd_setTriangles.vertexCount = length;
	cmd_setTriangles.vertexOffset = start;
	pushCommand(JSON.parse(JSON.stringify(cmd_setTriangles))); //deep-copy json parsable object
}

function ToHex(d) {

    var sign = "0";

    if(d<0.0){
        sign = "1";
        d = -d;
    }

    var mantissa = parseFloat(d).toString(2);

    var exponent = 0;

    if(mantissa.substr(0,1) === "0"){
        exponent = mantissa.indexOf('.') - mantissa.indexOf('1') + 127;
    }
    else{
        exponent = mantissa.indexOf('.') - 1 + 127;
    }

    mantissa = mantissa.replace(".", "");
    mantissa = mantissa.substr(mantissa.indexOf('1')+1);

    if(mantissa.length>23){
        mantissa = mantissa.substr(0,23);
    }
    else{
        while(mantissa.length<23){
            mantissa = mantissa +"0";
        }
    }

    var exp = parseFloat(exponent).toString(2);

    while(exp.length<8){
        exp = "0" + exp;
    }

    var numberFull = sign + exp + mantissa;

    return parseInt(numberFull, 2).toString(16);
}

function sendModel(model){
	//this has to be multiple of 9 (because 9 vertex floats represent one triangle). 
	//also the entire cmd message (including header) has to be shorter then 65.535 (int length) characters
	// -> 8100*8 = 64.800 seems a perfect fit and leaves some room for the header and rest of the command.
	var packageSize = 8100; 
	var vertexCount = model.geometry.attributes.position.array.length;
	for(var i = 0; i<vertexCount;i+=packageSize)
	{
		setTriangles(model.geometry.attributes.position.array,i,Math.min(packageSize,vertexCount-i));
	}
	pushCommand(cmd_buildModel);
}

//makes sure all vertices are in the range [-0.5,0.5]
function noramlizeModel(model){
	modelSize = model.geometry.attributes.position.array.length;
	var x,y,z;
	var minX = 1000000.0,maxX = -1000000.0;
	var minY = 1000000.0,maxY = -1000000.0;
	var minZ = 1000000.0,maxZ = -1000000.0;
	for(var i = 0; i<model.geometry.attributes.position.array.length;i+=3)
	{
		x = model.geometry.attributes.position.array[i];
		y = model.geometry.attributes.position.array[i+1];
		z = model.geometry.attributes.position.array[i+2];
		
		if(x<minX)
			minX = x;
		if(y<minY)
			minY = y;
		if(z<minZ)
			minZ = z;
		
		if(x>maxX)
			maxX = x;
		if(y>maxY)
			maxY = y;
		if(z>maxZ)
			maxZ = z;			
	}
	
	var maxRange = maxX-minX;
	if(maxY-minY>maxRange)
		maxRange = maxY-minY;
	if(maxZ-minZ>maxRange)
		maxRange = maxZ-minZ;
	
	for(var i = 0; i<model.geometry.attributes.position.array.length;i+=3)
	{
		x = model.geometry.attributes.position.array[i];
		y = model.geometry.attributes.position.array[i+1];
		z = model.geometry.attributes.position.array[i+2];
		
		x = (x-(maxX+minX)*0.5)/maxRange;
		y = (y-(maxY+minY)*0.5)/maxRange;
		z = (z-(maxZ+minZ)*0.5)/maxRange;
		
		model.geometry.attributes.position.array[i] = x;
		model.geometry.attributes.position.array[i+1] = y;
		model.geometry.attributes.position.array[i+2] = z;
	}
	
	model.geometry.attributes.position.needsUpdate = true;
}

function createProjection(){
	pushCommand(cmd_createProjection);
}