var cmd_getPosition = {cmd : 'getPosition'};
var cmd_getButtons = {cmd : 'getButtons'};
var cmd_ping = {cmd : 'ping'};
var cmd_getConnection = {cmd : 'getConnection'};
var cmd_getCollision = {cmd : 'getCollision'};
var cmd_getClosestPoint = {cmd : 'getClosestPoint'};
var cmd_getCurvePoint = {cmd : 'getCurvePoint'};
var cmd_clearModel = {cmd : 'clearModel'};
var cmd_buildModel = {cmd : 'buildModel'};
var cmd_setTriangle = {cmd : 'setTriangle', id : 0, ax : 0.0, ay : 0.0, az : 0.0, bx : 0.0, by : 0.0, bz : 0.0, cx : 0.0, cy : 0.0, cz : 0.0};
var cmd_setTriangles = {cmd : 'setTriangles', vertexOffset : 0, vertexCount : 0, vertexArray : ''};
var cmd_removeTriangle = {cmd : 'removeTriangle', id : 0};
var cmd_setCursorSize = {cmd : 'setCursorSize', size : 0.06};
var cmd_setHapticMode = {cmd : 'setHapticMode', mode : 0};
var cmd_setMagnetic = {cmd : 'setMagnetic', magnetic : 0};

var cmd_setProjectionVector = {cmd: 'setProjectionVector', px: 0.0, py: 0.0, pz: 1.0};
var cmd_setPathSpeed = {cmd: 'setPathSpeed', speed : 0.3};

var cmd_setExternalCursor = {cmd : 'setExternalCursor', externalX : 0.0, externalY : 0.0, externalZ : 0.0, active : 0};

var previousPosition = {cmd : 'getPosition', x : 0.0, y: 0.0, z: 0.0};
var position = {cmd : 'getPosition', x : 0.0, y: 0.0, z: 0.0};
var buttons = {cmd : 'getButtons', button0 : 0, button1 : 0, button2 : 0, button3 : 0};
var pathDeviation = 0.0;
var curvePoint = {cmd : 'getCurvePoint', x: 0.0, y: 0.0, z: 0.0, index: 0};
var connection = -1;
var cursorSize_defaultMode0 = 0.06;
var cursorSize_defaultMode1 = 0.02;
var cursorSize = cursorSize_defaultMode0;
var pathSpeed = 0.4;
var hapticMode = 0;

var collision = false;
var closestPoint = {cmd : 'getClosestPoint', x : 0.0, y: 0.0, z: 0.0};
var modelSize = 0;
var lastLoad = 0;
var loading  = 0.0;
var buildingModel = false;
var settingProjectionVector = false;

var cmdArray = [];
  
function pushCommand(cmdObj){
	if(cmdObj.cmd == "buildModel"){
		buildingModel = true;
	}
	if(cmdObj.cmd == "setProjectionVector"){
		settingProjectionVector = true;
	}
	cmdArray.push(cmdObj);
	return true;
}

function processResult(result){
	switch(result.cmd){
		case "debug": //debug and error messages should be handled in the contentScript
			break;
		case "error":
			break;
		case "getPosition":
			previousPosition = position;
			position = result;
			break;
		case "getButtons":
			buttons = result;
			break;
		case "getConnection":
			connection = result.connection;
			break;
		case "getCollision":
			collision = result.collision;
			break;
		case "getClosestPoint":
			closestPoint = result;
			break;
		case "getCurvePoint":
			curvePoint = result;
			break;
		case "setTriangle":
			if(result.id-lastLoad>10000){
				loading = result.id/modelSize;
				lastLoad = result.id;
				percentComplete = loading;
				console.log("loading model: "+Math.floor(100*loading)+"%");
			}
			break;
		case "setTriangles":
			if(result.id-lastLoad>10000){
				loading = result.id/modelSize;
				lastLoad = result.id;
				percentComplete = loading;
				console.log("loading model: "+Math.floor(100*loading)+"%");
			}
			break;
		case "removeTriangle":
			break;
		case "setProjectionVector":
			console.log(result);
			settingProjectionVector = false;
			break;
		case "clearModel":
			console.log(result);
			break;
		case "buildModel":
			console.log('model is ready for haptics.');
			buildingModel = false;
			break;
	}
}

function processCommandArray(){
	var returnString = document.getElementById('fd_return').innerHTML;
	if(returnString!="" && returnString != "[]"){
		var result = JSON.parse(returnString);

		for(let i = 0; i<result.length;++i){
			processResult(result[i]);
		}
	}
	document.getElementById('fd_return').innerHTML = "[]";

	var cmdElement = document.getElementById('fd_command')

	if(cmdArray.length==0)
		return;

	let shiftElements = 0;
	for(let i = 0; i<cmdArray.length;++i){
		var elementString = JSON.stringify(cmdArray[i]);
		if(cmdElement.innerHTML.length+elementString.length>60000)
			break;

		if(cmdElement.innerHTML.length < 3)
		{
			cmdElement.innerHTML = "["+elementString+"]";
		}
		else{
			cmdElement.innerHTML = cmdElement.innerHTML.substr(0,cmdElement.innerHTML.length-1)+  "," + elementString+"]";
		}
		++shiftElements;
	}

	for(let i = 0;i<shiftElements;++i)
		cmdArray.shift();
}

function refreshPositionAndButtons(){
	if(buildingModel || settingProjectionVector)
		return;
	if(connection>=0 && cmdArray.length<32){
		pushCommand(cmd_getPosition);
		pushCommand(cmd_getButtons);
		pushCommand(cmd_getCollision);
		pushCommand(cmd_getClosestPoint);
		pushCommand(cmd_getCurvePoint);
		pushCommand(cmd_setExternalCursor);
	}
}

//TIRANGLE AND MODEL SYNCHRONIZATION

function setTriangle(id, ax, ay, az, bx, by, bz, cx, cy, cz){
	cmd_setTriangle.id = id;
	
	cmd_setTriangle.ax = ax;
	cmd_setTriangle.ay = ay;
	cmd_setTriangle.az = az;
	
	cmd_setTriangle.bx = bx;
	cmd_setTriangle.by = by;
	cmd_setTriangle.bz = bz;
	
	cmd_setTriangle.cx = cx;
	cmd_setTriangle.cy = cy;
	cmd_setTriangle.cz = cz;

	pushCommand(JSON.parse(JSON.stringify(cmd_setTriangle))); //deep-copy json parsable object
}


function setTriangles(vertexArray, start, length){
	triangles = new Array(length);
	triangleHexString = "";
	
	for(var i = start; i<start+length;++i)
	{
		triangleHexString+=ToHex(vertexArray[i]);
	}
	
	cmd_setTriangles.vertexArray = triangleHexString;
	cmd_setTriangles.vertexCount = length;
	cmd_setTriangles.vertexOffset = start;
	pushCommand(JSON.parse(JSON.stringify(cmd_setTriangles))); //deep-copy json parsable object
}

var ToHexDataView = new DataView(new ArrayBuffer(4)),result;
function ToHex(f){
	const getHex = i => ('00' + i.toString(16)).slice(-2);

	ToHexDataView.setFloat32(0, f);

	result = Array
		.apply(null, { length: 4 })
		.map((_, i) => getHex(ToHexDataView.getUint8(i)))
		.join('');

	return result;
}

function sendModel(model){	
	//this has to be multiple of 9 (because 9 vertex floats represent one triangle). 
	//also the entire cmd message (including header) has to be shorter then 65.535 (int length) characters
	// -> 8100*8 = 64.800 seems a perfect fit and leaves some room for the header and rest of the command.

	var packageSize = 4500; 
	var vertexCount = model.geometry.attributes.position.array.length;
	for(var i = 0; i<vertexCount;i+=packageSize)
	{
		setTriangles(model.geometry.attributes.position.array,i,Math.min(packageSize,vertexCount-i));
	}
	pushCommand(cmd_buildModel);
}

//makes sure all vertices are in the range [-0.5,0.5]
function noramlizeModel(model){
	modelSize = model.geometry.attributes.position.array.length;
	var x,y,z;
	var minX = 1000000.0,maxX = -1000000.0;
	var minY = 1000000.0,maxY = -1000000.0;
	var minZ = 1000000.0,maxZ = -1000000.0;
	for(var i = 0; i<model.geometry.attributes.position.array.length;i+=3)
	{
		x = model.geometry.attributes.position.array[i];
		y = model.geometry.attributes.position.array[i+1];
		z = model.geometry.attributes.position.array[i+2];
		
		if(x<minX)
			minX = x;
		if(y<minY)
			minY = y;
		if(z<minZ)
			minZ = z;
		
		if(x>maxX)
			maxX = x;
		if(y>maxY)
			maxY = y;
		if(z>maxZ)
			maxZ = z;			
	}
	
	var maxRange = maxX-minX;
	if(maxY-minY>maxRange)
		maxRange = maxY-minY;
	if(maxZ-minZ>maxRange)
		maxRange = maxZ-minZ;
	
	for(var i = 0; i<model.geometry.attributes.position.array.length;i+=3)
	{
		x = model.geometry.attributes.position.array[i];
		y = model.geometry.attributes.position.array[i+1];
		z = model.geometry.attributes.position.array[i+2];
		
		x = (x-(maxX+minX)*0.5)/maxRange;
		y = (y-(maxY+minY)*0.5)/maxRange;
		z = (z-(maxZ+minZ)*0.5)/maxRange;
		
		model.geometry.attributes.position.array[i] = x;
		model.geometry.attributes.position.array[i+1] = y;
		model.geometry.attributes.position.array[i+2] = z;
	}
	
	model.geometry.attributes.position.needsUpdate = true;
}

//makes sure all vertices are in the range [-0.5,0.5]
function rotateModel(model, radx, rady, radz){
	modelSize = model.geometry.attributes.position.array.length;
	
	var rotationX = new THREE.Matrix4().makeRotationX(radx);
	var rotationY = new THREE.Matrix4().makeRotationY(rady);
	var rotationZ = new THREE.Matrix4().makeRotationZ(radz);
	
	var rotation = rotationX.multiply(rotationY.multiply(rotationZ));
	
	model.geometry.applyMatrix(rotation);
	
	model.geometry.attributes.position.needsUpdate = true;
}

function setProjectionVector(x,y,z){
	cmd_setProjectionVector.px = x;
	cmd_setProjectionVector.py = y;
	cmd_setProjectionVector.pz = z;
	pushCommand(cmd_setProjectionVector);
}

function setPathSpeed(speed){
	cmd_setPathSpeed.speed = speed;
	pushCommand(cmd_setPathSpeed);
}

function setHapticsMode(hapticMode){
	cmd_setHapticMode.mode = hapticMode;
	pushCommand(cmd_setHapticMode);
}

function setMagnetic(magnetic){
	cmd_setMagnetic.magnetic = magnetic;
	pushCommand(cmd_setMagnetic);
}

function setCursorSize(cursorSize){
	cmd_setCursorSize.size = cursorSize;
	pushCommand(cmd_setCursorSize);
}

function setExternalCursor(x,y,z,active){
	cmd_setExternalCursor.externalX = x;
	cmd_setExternalCursor.externalY = y;
	cmd_setExternalCursor.externalZ = z;
	cmd_setExternalCursor.active = active;
}

function resetHaptics(){
	pushCommand(cmd_getConnection);
	pushCommand(cmd_clearModel);
	cmd_setHapticMode.mode = 0;
	pushCommand(cmd_setHapticMode);
}

setInterval(processCommandArray,6);
setInterval(refreshPositionAndButtons,20);
resetHaptics();