if ( WEBGL.isWebGLAvailable() === false ) {

	document.body.appendChild( WEBGL.getWebGLErrorMessage() );

}

var container, stats;

var camera = null, scene, renderer;
var cameraControls;
var effectController;
var teapotSize = 400;
var ambientLight, light;

var tess = -1;	// force initialization
var bBottom ;
var bLid;
var bBody;
var bFitLid;
var bNonBlinn;
var modelName = "";
var shading;
var cameraMode;
var wireMaterial, smoothMaterial, flatMaterial;

var model = null;

var codeArea = null;

// allocate these just once
var diffuseColor = new THREE.Color();
var specularColor = new THREE.Color();

var cursorGizmo = null;
var closestPointGizmo;

var sound;
var touchSound;
var hitSound;
var curveSound;
var blingSound;
var usePitchShift;
var audioListener;

var raycaster = new THREE.Raycaster();
var mouse = new THREE.Vector2();
var mouseWheelDelta = 0;
var hoverObject = null;
var draggingCursor = false;
var draggingZ;

var hideModel = false;

var autoFileName = "../../test.stl";
var autoFileHash = 0;

init();
animate();

var loadingUI;
function init() {

	container = document.getElementById( 'container' );
	
	loadingUI =  document.getElementById( 'loading' );
	
	codeArea = document.getElementById( 'codeArea' );
	debugArea = document.getElementById( 'debugArea' );

	var canvasWidth = container.offsetWidth;
	var canvasHeight = container.offsetHeight;

	console.log(document.getElementsByClassName( 'resizable' ));
	//document.getElementsByClassName( 'resizable' ).handle = document.getElementById( 'handle' );

	//AUDIO
	
	audioListener = new THREE.AudioListener(); //add audio listener
	var audioLoader = new THREE.AudioLoader();
	
	
	var refDistance = 1;
	var rollOff = 1;
	var volume = 0.0;
	
	touchSound = new THREE.PositionalAudio( audioListener );
		audioLoader.load( 'sounds/deepnoise.ogg', function ( buffer ) {
			touchSound.setBuffer( buffer );
			touchSound.setRefDistance( refDistance );
			touchSound.setRolloffFactor( rollOff );
			touchSound.setLoop( true );
			touchSound.setVolume( volume );
			touchSound.panner.panningModel = 'HRTF';
			touchSound.setDistanceModel('inverse');
			touchSound.play();			
		} );
	
	hitSound = new THREE.PositionalAudio( audioListener );
		audioLoader.load( 'sounds/hit.ogg', function ( buffer ) {
			hitSound.setBuffer( buffer );
			touchSound.setRefDistance( refDistance );
			touchSound.setRolloffFactor( rollOff );
			hitSound.setLoop( false );
			hitSound.setVolume( volume );	
			hitSound.panner.panningModel = 'HRTF';
			hitSound.setDistanceModel('inverse');
			hitSound.play();
		} );

	curveSound = new THREE.Audio( audioListener );
	audioLoader.load( 'sounds/220.wav', function ( buffer ) {
		curveSound.setBuffer( buffer );
		curveSound.setLoop( true );
		curveSound.setVolume( volume );
		curveSound.play();		
	} );

	blingSound = new THREE.Audio( audioListener );
		audioLoader.load( 'sounds/bling220.wav', function ( buffer ) {
			blingSound.setBuffer( buffer );
			blingSound.setLoop( false );
			blingSound.setVolume( 1 );	
	} );
	
	createPositionalAudio('sounds/tone.ogg');
	
	// LIGHTS
	ambientLight = new THREE.AmbientLight( 0x333333 );	// 0.2

	light = new THREE.DirectionalLight( 0xFFFFFF, 1.0 );
	// direction is set in GUI

	// RENDERER
	renderer = new THREE.WebGLRenderer( { antialias: true } );
	renderer.setPixelRatio( window.devicePixelRatio );
	renderer.setSize( canvasWidth, canvasHeight );
	renderer.gammaInput = true;
	renderer.gammaOutput = true;


	container.appendChild( renderer.domElement );

	// EVENTS
	window.addEventListener( 'resize', onWindowResize, false );

	// MATERIALS
	var materialColor = new THREE.Color();
	materialColor.setRGB( 1.0, 1.0, 1.0 );

	smoothMaterial = new THREE.MeshPhongMaterial( { color: materialColor, specular: 0x000000, flatShading: false, side: THREE.DoubleSide } );
	flatMaterial = new THREE.MeshPhongMaterial( { color: materialColor, specular: 0x000000, flatShading: true, side: THREE.DoubleSide } );
	wireMaterial = new THREE.MeshBasicMaterial( { color: 0xFFFFFF, wireframe: true } ) ;
	

	
	// scene itself
	scene = new THREE.Scene();
	scene.background = new THREE.Color( 0x111111 );
	// scene.background = new THREE.Color( 0xFFFFFF );

	scene.add( ambientLight );
	scene.add( light );

	// 3D OBJ MODEL
	//loaded on UI change
	
	// CAMERA
	refreshCamera();
	
	//CURSOR
	refreshCursor();
	
	//CLOSEST POINT
	var geometry = new THREE.SphereGeometry( 0.02, 32, 32 );
	var material = new THREE.MeshBasicMaterial( {color: 0x00FF00} );
	closestPointGizmo = new THREE.Mesh( geometry, material );
	scene.add(closestPointGizmo);
	
	// GUI
	setupGui();
	
	stats = new Stats();
	//container.appendChild( stats.dom );

	window.addEventListener('focus', checkAutoFile);
	checkAutoFile();
}

function loadModelFromSTLString(stlString){
	setupHapticsMode(0);
	if(model!=null)
		scene.remove(model);
	resetHaptics();

	fileLoader = new THREE.STLLoader();
	geometry = fileLoader.parse(stlString);
	var material = flatMaterial;
	var mesh = new THREE.Mesh( geometry, material )
	noramlizeModel(mesh);
	rotateModel(mesh,-Math.PI*0.5,0,0);
	mesh.rotation.set( 0, 0, 0 );
	mesh.scale.set( 1, 1, 1 );
	model = mesh;
	model.userData.type = "model";
	scene.add( model );
	sendModel(model);
	setProjectionVector(0,0,1);
}

function loadModel(modelFile, fileData = undefined){
	var fileExtension = modelFile.substr(-3);
	if(fileData==undefined)
		fileData = modelFile;

	setupHapticsMode(0);
	if(model!=null)
		scene.remove(model);
	resetHaptics();
	var fileLoader;


	if(fileExtension === 'obj')	
	{
		fileLoader = new THREE.OBJLoader();
		fileLoader.load(fileData, function (object) {
		model = object.children[0];
		noramlizeModel(model);
		model.traverse(function(node) {
			node.material = flatMaterial;
		});
		model.userData.type = "model";
		scene.add( model);
		sendModel(model);
		setProjectionVector(0,0,1);
    });
	}
	else if(fileExtension === 'stl')
	{
		fileLoader = new THREE.STLLoader();
		fileLoader.load(fileData, function (geometry) {
			var material = flatMaterial;
			var mesh = new THREE.Mesh( geometry, material )
			noramlizeModel(mesh);
			rotateModel(mesh,-Math.PI*0.5,0,0);
			mesh.rotation.set( 0, 0, 0 );
			mesh.scale.set( 1, 1, 1 );
			model = mesh;
			model.userData.type = "model";
			scene.add( model );
			sendModel(model);
			setProjectionVector(0,0,1);
		});
	}
	else
	{
		console.log("file extension not supported: "+fileExtension);
	}
}


var gui;
function setupGui() {

	effectController = {
		hue:		0.45,
		saturation: 1.00,
		lightness:  0.3,

		lhue:		 0.04,
		lsaturation: 0.01,	// non-zero so that fractions will be shown
		llightness:  1.0,

		// bizarrely, if you initialize these with negative numbers, the sliders
		// will not show any decimal places.
		lx: 0.32,
		ly: 0.39,
		lz: 0.7,
		newTess: 15,
		bottom: true,
		lid: true,
		body: true,
		fitLid: false,
		nonblinn: false,
		model: "study/cube.obj",
		newShading: "flat",
		cameraMode: "perspective"
	};

	var h;
	gui = new dat.GUI();
	// material (color)

	h = gui.addFolder( "Material color" );

	h.add( effectController, "hue", 0.0, 1.0, 0.025 ).name( "hue" );
	h.add( effectController, "saturation", 0.0, 1.0, 0.025 ).name( "saturation" );
	h.add( effectController, "lightness", 0.0, 1.0, 0.025 ).name( "lightness" );

	// light (point)

	h = gui.addFolder( "Lighting" );

	h.add( effectController, "lhue", 0.0, 1.0, 0.025 ).name( "hue" );
	h.add( effectController, "lsaturation", 0.0, 1.0, 0.025 ).name( "saturation" );
	h.add( effectController, "llightness", 0.0, 1.0, 0.025 ).name( "lightness" );

	// light (directional)

	h = gui.addFolder( "Light direction" );

	h.add( effectController, "lx", -1.0, 1.0, 0.025 ).name( "x" );
	h.add( effectController, "ly", -1.0, 1.0, 0.025 ).name( "y" );
	h.add( effectController, "lz", -1.0, 1.0, 0.025 ).name( "z" );

	// shading and model

	gui.add( effectController, "model", ["study/cube.obj","study/sphere.obj","study/bunny.obj","study/1.obj","study/2.obj","study/3.obj","study/4.obj","study/5.obj","study/6.obj"] ).name( "Model" );
	
	gui.add( effectController, "newShading", [ "flat","smooth","wireframe"] ).name( "Shading" );
	
	gui.add( effectController, "cameraMode", [ "perspective","orthographic"] ).name( "Camera" );


	soundController = {
		masterVolume: 1.0,
		usePitchShift: "Y"
	};
	h = gui.addFolder( "Sound" );
	h.add( soundController, "masterVolume", 0.0, 1.0, 0.01 ).name( "Volume" );
	h.add( soundController, "usePitchShift",["X","Y","Z","Distance","Off"]).name( "Pitch Shift" );

	gui.close();
	//dat.GUI.toggleHide();
}

function showLoading(){
	loadingUI.style.display = "block";
}

function hideLoading(){
	loadingUI.style.display = "none";
}

function setupHapticsMode(mode){
	hapticMode = mode;
		
	switch(hapticMode){
		case 0:
			cursorSize = cursorSize_defaultMode0;
			effectController.cameraMode = "perspective";
			break;
		case 1:
			cursorSize = cursorSize_defaultMode1;
			effectController.cameraMode = "orthographic";
			if(cmd_setProjectionVector.z<0.9)
				setProjectionVector(0,0,1);
			break;
	}
	
	refreshCamera();
	refreshCursor();
	refreshGUI();
	setHapticsMode(hapticMode);
}

function animate() {
	if((buildingModel || settingProjectionVector) && connection>=0)
		showLoading();
	else
		hideLoading();
	requestAnimationFrame( animate );
	render();
	stats.update();
	updateInputs();
	updateCursor();
	updateSound();
}

function render() {

	if ( effectController.newTess !== tess ||
		effectController.bottom !== bBottom ||
		effectController.lid !== bLid ||
		effectController.body !== bBody ||
		effectController.fitLid !== bFitLid ||
		effectController.nonblinn !== bNonBlinn ||
		effectController.newShading !== shading  ||
		effectController.cameraMode !== cameraMode)	
	{

		tess = effectController.newTess;
		bBottom = effectController.bottom;
		bLid = effectController.lid;
		bBody = effectController.body;
		bFitLid = effectController.fitLid;
		bNonBlinn = effectController.nonblinn;
		shading = effectController.newShading;
		cameraMode = effectController.cameraMode;
		
		refreshMaterial();
		refreshCamera();

	}
	
	if(effectController.model !== modelName){
		modelName = effectController.model;
		if(modelName!="")
			loadModel('models/'+modelName);
	}


	diffuseColor.setHSL( effectController.hue, effectController.saturation, effectController.lightness );

	smoothMaterial.color.copy( diffuseColor );
	flatMaterial.color.copy( diffuseColor );

	// Ambient's actually controlled by the light for this demo
	ambientLight.color.setHSL( effectController.hue, effectController.saturation, effectController.lightness);

	light.position.set( effectController.lx, effectController.ly, effectController.lz );
	light.color.setHSL( effectController.lhue, effectController.lsaturation, effectController.llightness );


	if(hideModel){
		var clearColor = renderer.getClearColor();
		renderer.setClearColor(new THREE.Color(0x444444));
		renderer.clearColor();
		renderer.setClearColor(clearColor);
	}else{
		renderer.render( scene, camera );
	}

}


//REBUILD/REFRESH FUNCTIONS

// Whenever the teapot changes, the scene is rebuilt from scratch (not much to it).
function refreshMaterial() {
	if(!model)
		return;
	switch(shading){
		case "wireframe":
			model.traverse(function(node) {
				node.material = wireMaterial;
			});
			break;
		case "flat":
			model.traverse(function(node) {
				node.material = flatMaterial;
			});
			break;
		default:
			model.traverse(function(node) {
				node.material = smoothMaterial;
			});
			break;
		
	}
}

function refreshCamera(position = null){
	var pos = new THREE.Vector3(0,0,2.5);
	if(camera!=null){
		pos = camera.position;
		scene.remove(camera);
	}
	if(position != null){
		pos = position;
	}
	
	
	var canvasWidth = container.offsetWidth;
	var canvasHeight = container.offsetHeight;

	var frustumSize = 2;
	var aspect = canvasWidth / canvasHeight;
	camera = new THREE.PerspectiveCamera( 45, aspect, 1, 80000 );

	switch(cameraMode){
		case "perspective":
			camera = new THREE.PerspectiveCamera( 45, aspect, 1, 80000 );
			break;
		case "orthographic":
			camera = new THREE.OrthographicCamera( frustumSize * aspect / - 2, frustumSize * aspect / 2, frustumSize / 2, frustumSize / - 2, 1, 1000 );
			break;
	}
	camera.position.set( pos.x,pos.y,pos.z );
	scene.add(camera);
	
	cameraControls = new THREE.OrbitControls( camera, renderer.domElement );
	cameraControls.addEventListener( 'change', render );
}

function refreshCursor(){
	if(cursorGizmo!=null)
		scene.remove(cursorGizmo);
	var geometry = new THREE.SphereGeometry( cursorSize, 32, 32 );
	var material = new THREE.MeshBasicMaterial( {color: 0xFF0000} );
	cursorGizmo = new THREE.Mesh( geometry, material );
	cursorGizmo.userData.type = "cursor";
	scene.add(cursorGizmo);
	cursorGizmo.add(audioListener);
	setCursorSize(cursorSize);
}

function refreshGUI(){
	for (var i in gui.__controllers) {
		gui.__controllers[i].updateDisplay();
	}
	for (var i = 0; i < Object.keys(gui.__folders).length; i++) {
		var key = Object.keys(gui.__folders)[i];
		for (var j = 0; j < gui.__folders[key].__controllers.length; j++ )
		{
			gui.__folders[key].__controllers[j].updateDisplay();
		}
	}
}


//UPDATE FUNCTIONS

function updateCursor(){
	if(connection>=0){
		cursorGizmo.position.set(position.x,position.y,position.z);
		closestPointGizmo.position.set(closestPoint.x,closestPoint.y,closestPoint.z)
		if(collision == 0)
			closestPointGizmo.material.color.setHex(0x00FF00);
		else
			closestPointGizmo.material.color.setHex(0xFF0000);
	}else{
		cursorGizmo.position.set(100000,0.0,0.0);
		closestPointGizmo.position.set(100000,0.0,0.0);
	}
}

var curveIndexLastUpdate = 0;
var semiTones = [0,2,4,5,7,9,11,
				12,14,16,17,19,21,23,
				24,26,28,29,31,33,35,
				36,35,33,31,29,28,26,
				24,23,21,19,17,16,14,
				12,11,9,7,5,4,2,
				0,4,7,
				12,7,4];
var collisionLastFrame = 0;

function updateSound(){
	if(position.x == 0 && position.y == 0 && position.z == 0)
		return;
	
	var distance = cursorGizmo.position.length();
	
	curveSound.setVolume(curveSound.getVolume()*0.95);
	if(hapticMode==1 && curvePoint.index!=curveIndexLastUpdate && usePitchShift!= "Off"){
		var pitch;

		if(usePitchShift == "Distance")
			pitch = (1.0*distance+0.5)*10.0;
		else if(usePitchShift == "X")
			pitch = (1.0*curvePoint.x+0.5)*10.0;
		else if(usePitchShift == "Y")
			pitch = (1.0*curvePoint.y+0.5)*10.0;
		else if(usePitchShift == "Z")
			pitch = (-1.0*curvePoint.z+0.5)*10.0;
		else
			pitch = curvePoint.index;
		
		console.log(curvePoint.index);

		if(pitch<0) pitch= 0;
		
		if(curvePoint.index>=0){
			curveSound.setPlaybackRate(Math.pow(Math.pow(2,1.0/12),semiTones[Math.round(pitch)%semiTones.length]));
			curveSound.setVolume(1.0);
		}
		if(curvePoint.index == 0){
			//blingSound.setPlaybackRate(Math.pow(Math.pow(2,1.0/12),semiTones[Math.round(pitch)%semiTones.length]));
			blingSound.play();
		}
		curveIndexLastUpdate = curvePoint.index;
	}
	
	//TODO: adjust audio when switching modes
	if(hapticMode==1){
		updatePannerPosition(-cursorGizmo.position.x,10);
	}
	else
		updatePannerPosition(-cursorGizmo.position.x,closestPointGizmo.position.distanceTo(cursorGizmo.position));
	
	if(collision>0){
		var diffX = (position.x-previousPosition.x);
		var diffY = (position.y-previousPosition.y);
		var diffZ = (position.z-previousPosition.z);
		var d = Math.sqrt(diffX*diffX + diffY*diffY + diffZ*diffZ);
		touchSound.setVolume(Math.min(1.0,d*0.7));
		if(collisionLastFrame<=0){
			hitSound.setVolume(Math.min(1.0,Math.pow(d*2,2)));
			hitSound.stop();
			hitSound.play();
		}
	}
	else
		touchSound.setVolume(0);
	
	
	collisionLastFrame = collision;
	
	audioListener.setMasterVolume(soundController.masterVolume);
	usePitchShift = soundController.usePitchShift;
}

var buttonsLastFrame = JSON.parse(JSON.stringify(buttons));
function updateInputs(){
	//switching modes
	if(buttons.button0>0 && buttonsLastFrame.button0==0){
		hapticMode++;
		if(hapticMode>1)
			hapticMode=0;		
		setupHapticsMode(hapticMode);
		refreshCursor();
	}
	if(buttonsLastFrame== undefined)
		buttonsLastFrame = JSON.parse(JSON.stringify(buttons));

	//control mode
	switch(hapticMode){
		case 0:
			if(buttons.button1>0){
				cursorSize = Math.max(0.03,cursorSize-0.005);
				refreshCursor();
			}
			if(buttons.button3>0){
				cursorSize = Math.min(0.1,cursorSize+0.005);
				refreshCursor();
			}
			if(buttons.button2>0 && buttonsLastFrame.button2==0){
				setMagnetic(1);
			}
			if(buttons.button2==0 && buttonsLastFrame.button2>0){
				setMagnetic(0);
			}
			break;
		case 1:
			if(buttons.button1>0 && buttonsLastFrame.button1==0){
			  setProjectionVector(1,0,0);
			  soundController.usePitchShift = "Y";
			  refreshGUI();
			}
			if(buttons.button2>0 && buttonsLastFrame.button2==0){
			  setProjectionVector(0,1,0);
			  soundController.usePitchShift = "Z";
			  refreshGUI();
			}
			if(buttons.button3>0 && buttonsLastFrame.button3==0){
			  setProjectionVector(0,0,1);
			  soundController.usePitchShift = "Y";
			  refreshGUI();
			}
			break;
	}
	
	updateMouseToCursorControl();
	
	buttonsLastFrame = JSON.parse(JSON.stringify(buttons))
}

function updateMouseToCursorControl(){
	if(draggingCursor){
		var pos = new THREE.Vector3( mouse.x, mouse.y, draggingZ ).unproject( camera );
		cursorGizmo.position.set(pos.x,pos.y,pos.z);
		setExternalCursor(pos.x,pos.y,pos.z,1);
		return;
	}
	
	
	raycaster.setFromCamera( mouse, camera );
	var intersects = raycaster.intersectObjects( scene.children );
	if(intersects.length>0)
		hoverObject =intersects[ 0 ];
	else
		hoverObject = null
	

}



// EVENT HANDLERS

function onWindowResize() {
	var canvasWidth = container.offsetWidth;
	var canvasHeight = container.offsetHeight;

	renderer.setSize( canvasWidth, canvasHeight );

	camera.aspect = canvasWidth / canvasHeight;
	camera.updateProjectionMatrix();

	render();
}

function handleDragOver(evt) {
    evt.stopPropagation();
    evt.preventDefault();
    evt.dataTransfer.dropEffect = 'copy'; 
}

function getHash(obj){
    var hash = 0;
    for (var i = 0; i < obj.length; i++) {
        var character = obj.charCodeAt(i);
        hash = ((hash<<5)-hash)+character;
        hash = hash & hash; // Convert to 32bit integer
    }
    return hash;
}

function checkAutoFile(){
	var loader = new THREE.FileLoader();


	//load a text file and output the result to the console
	loader.load(
		// resource URL
		autoFileName,
	
		// onLoad callback
		function ( data ) {
			console.log("reloading auto file...");
			// output the text to the console
			effectController.model = "";
			modelName ="";

			hash = getHash(data);
			if(hash!=autoFileHash){
				autoFileHash = hash;
				loadModel(autoFileName);
			}
		},
	
		// onError callback
		function ( err ) {

			console.log("error while loading file ["+autoFileName+"]");
			console.log(err);
		}
	);

}

function dropFile(evt) {
    evt.stopPropagation();
    evt.preventDefault();

    var file = evt.dataTransfer.files[0]; // FileList Objekt

	var reader = new FileReader();

	var senddata = new Object();
    // Read Metadata
    senddata.name = file.name;
    senddata.date = file.lastModified;
    senddata.size = file.size;
    senddata.type = file.type;

    // Wenn der Dateiinhalt ausgelesen wurde...
    reader.onload = function(theFileData) {
      senddata.fileData = theFileData.target.result; // Ergebnis vom FileReader auslesen
	  console.log(senddata);
      /*
      Code for AJAX-Requests
      */
	 loadModel(senddata.name,senddata.fileData);
    }

    // Convert file to URL
	reader.readAsDataURL(file);
  }

  function download(filename, text) {
	var element = document.createElement('a');
	element.setAttribute('href', 'data:text/plain;charset=utf-8,' + encodeURIComponent(text));
	element.setAttribute('download', filename);
  
	element.style.display = 'none';
	document.body.appendChild(element);
  
	element.click();
  
	document.body.removeChild(element);
  }
  
  function downloadModel(){
		  showLoading()
		  debugArea.value = "building...\n";
		  try {
  
			  var scriptText = codeEditor.getValue();
			  var scadModel = eval(scriptText);
			  var stlString = scadModel.toStlString();
			  loadModelFromSTLString(stlString);
			  download("model.stl",stlString);
			  debugArea.value += "object build successfully.";
			  debugArea.value = "downloading model...\n";
		  }catch(e){
			  console.log(e.stack);
  
			  let start = e.stack.indexOf("<anonymous>:")+12
			  let substr =  e.stack.substr(e.stack.indexOf("<anonymous>:")+12);
			  let line = substr.substr(0,substr.indexOf(')'));
			  if(start!=11)
				  debugArea.value = "ERROR at "+ line+ " : "+ e.message;
			  else
				  debugArea.value = "ERROR : "+ e.message;
				  debugArea.focus();
		  }
  }
  
window.addEventListener("keydown", function (event) {
	switch (event.keyCode) {		
		case 115: // F4:
		if(!codeAreaActivated)
			activateCodeArea(true);
		showLoading()
		debugArea.value = "building...\n";
		try {

			var scriptText = codeEditor.getValue();
			var scadModel = eval(scriptText);
			var stlString = scadModel.toStlString();
			loadModelFromSTLString(stlString)
			debugArea.value += "object build successfully.";
			container.focus();
		}catch(e){
			console.log(e.stack);

			let start = e.stack.indexOf("<anonymous>:")+12
			let substr =  e.stack.substr(e.stack.indexOf("<anonymous>:")+12);
			let line = substr.substr(0,substr.indexOf(')'));
			if(start!=11)
				debugArea.value = "ERROR at "+ line+ " : "+ e.message;
			else
				debugArea.value = "ERROR : "+ e.message;
				debugArea.focus();
		}
		break;
		case 27: //ESC
		codeEditor.focus();
		break;
	}

  if(codeEditorIsFocused){
		return;
	}

  if (event.defaultPrevented) {
    return; // Do nothing if the event was already processed
  }

  switch (event.keyCode) {
	case 32: // spacebar
		if(cmd_setMagnetic.magnetic==0)
			setMagnetic(1);
		break;
	case 49: // code for "1" key press.
	  setupHapticsMode(0);
	  break;
	case 50: // code for "2" key press.
	  setupHapticsMode(1);
	  break;
	case 88:      // code for "x" key press.
	  setProjectionVector(1,0,0);
	  refreshCamera(new THREE.Vector3(2.5,0,0));
	  soundController.usePitchShift = "Y";
	  refreshGUI();
	  break;
	case 89:      // code for "y" key press.
	  setProjectionVector(0,1,0);
	  refreshCamera(new THREE.Vector3(0,2.5,0));
	  soundController.usePitchShift = "Z";
	  refreshGUI();
	  break;
	case 90:      // code for "z" key press.
	  setProjectionVector(0,0,1);
	  refreshCamera(new THREE.Vector3(0,0,2.5));
	  soundController.usePitchShift = "Y";
	  refreshGUI();
		break;
	case 72:      // code for "h" key press.
		hideModel = !hideModel;
		break;
	case 107: //numpadd + 
		setPathSpeed(Math.min(cmd_setPathSpeed.speed+0.1,1));
		console.log("+");
		break;
	case 109: //numpad -
		console.log("-");
		setPathSpeed(Math.max(cmd_setPathSpeed.speed-0.1,0.1));
		console.log("-");
		break;

	default:
	  return; // Quit when this doesn't handle the key event.
  }

  // Cancel the default action to avoid it being handled twice
  event.preventDefault();
}, true);

window.addEventListener("keyup", function (event) {
	if (event.defaultPrevented) {
		return; // Do nothing if the event was already processed
	}

	switch (event.keyCode) {
		case 32: // spacebar
			setMagnetic(0);
			break;
	}
	event.preventDefault();
},true);

window.addEventListener( 'mousemove', function( event ) {

	// calculate mouse position in normalized device coordinates
	// (-1 to +1) for both components



	var rect = container.getBoundingClientRect();

	mouse.x = (event.clientX-rect.left)/(rect.right-rect.left)* 2 - 1;
	mouse.y = -(event.clientY-rect.top)/(rect.bottom-rect.top)* 2 + 1; 

}, false );

window.addEventListener("mousedown", function( event ) {
	if(hoverObject!=null && hoverObject.object.userData.type == "cursor"){
		console.log("start dragging");
		draggingCursor = true;
		draggingZ = cursorGizmo.position.project(camera).z;
		cameraControls.enabled = false;
		cameraControls.saveState();
	}
});

window.addEventListener("mouseup", function( event ) {
	if(draggingCursor){
		console.log("stop dragging");
		draggingCursor = false;
		cameraControls.enabled = true;
		cameraControls.reset();
		setExternalCursor(0,0,0,0);
	}
});

window.onbeforeunload = function ()
 {
     return "";
 };

container.addEventListener("mousewheel", function( event ) { //TODO: somehow this event is not trigger on the container (three.js is consuming it?)
	draggingZ += event.deltaY*0.1;
	console.log(draggingZ);
});

container.addEventListener('dragover', handleDragOver, false);
container.addEventListener('drop', dropFile, false);

document.addEventListener("keydown", function(e) {
	if ((window.navigator.platform.match("Mac") ? e.metaKey : e.ctrlKey)  && e.keyCode == 83) {
	  e.preventDefault();
	  downloadModel();
	}
  }, false);

codeEditor.focus();